from typing import Optional
from ..nocodb import (
    NocoDBClient,
    NocoDBProject,
    AuthToken,
    WhereFilter,
)
from ..api import NocoDBAPI
from ..utils import get_query_params
from ..exceptions import NocoDBAPIError

import requests


class NocoDBRequestsClient(NocoDBClient):
    def __init__(self, auth_token: AuthToken, base_uri: str):
        self.__session = requests.Session()
        self.__session.headers.update(
            auth_token.get_header(),
        )
        self.__session.headers.update({"Content-Type": "application/json"})
        self.__api_info = NocoDBAPI(base_uri)

    def _request(self, method: str, url: str, *args, **kwargs):
        response = self.__session.request(method, url, *args, **kwargs)
        try:
            response.raise_for_status()
            response_json = response.json()
        except requests.exceptions.JSONDecodeError:
            response_json = None
        except requests.exceptions.HTTPError as http_error:
            raise NocoDBAPIError(
                message=str(http_error),
                status_code=http_error.response.status_code,
                response_json=response_json,
                response_text=response.text
            )

        return response

    def table_row_list(
        self,
        project: NocoDBProject,
        table: str,
        filter_obj: Optional[WhereFilter] = None,
        params: Optional[dict] = None,
    ) -> dict:
        return self._request(
            "GET",
            self.__api_info.get_table_uri(project, table),
            params=get_query_params(filter_obj, params),
        ).json()

    def table_row_create(self, project: NocoDBProject, table: str, body: dict) -> dict:
        return self._request(
            "POST", self.__api_info.get_table_uri(project, table), json=body
        ).json()

    def table_row_detail(self, project: NocoDBProject, table: str, row_id: int) -> dict:
        return self._request(
            "GET",
            self.__api_info.get_row_detail_uri(project, table, row_id),
        ).json()

    def table_row_update(
        self, project: NocoDBProject, table: str, row_id: int, body: dict
    ) -> dict:
        return self._request(
            "PATCH",
            self.__api_info.get_row_detail_uri(project, table, row_id),
            json=body,
        ).json()

    def table_row_delete(self, project: NocoDBProject, table: str, row_id: int) -> int:
        return self._request(
            "DELETE",
            self.__api_info.get_row_detail_uri(project, table, row_id),
        ).json()

    def table_count(
        self,
        project: NocoDBProject,
        table: str,
        filter_obj: Optional[WhereFilter] = None,
    ) -> dict:
        return self._request(
            "GET",
            self.__api_info.get_table_count_uri(project, table),
            params=get_query_params(filter_obj),
        ).json()

    def table_find_one(
        self,
        project: NocoDBProject,
        table: str,
        filter_obj: Optional[WhereFilter] = None,
        params: Optional[dict] = None,
    ) -> dict:
        return self._request(
            "GET",
            self.__api_info.get_table_find_one_uri(project, table),
            params=get_query_params(filter_obj, params),
        ).json()

    def table_row_nested_relations_list(
        self,
        project: NocoDBProject,
        table: str,
        relation_type: str,
        row_id: int,
        column_name: str,
    ) -> dict:
        return self._request(
            "GET",
            self.__api_info.get_nested_relations_rows_list_uri(
                project, table, relation_type, row_id, column_name
            ),
        ).json()

    def project_create(self, body):
        return self._request(
            "POST", self.__api_info.get_project_uri(), json=body
        ).json()




class BasicFilter:
    def __init__(self, column_name: str, filter_name: str, value: str):
        self.__column_name = column_name
        self.__value = value
        self.__filter_name = filter_name

    def get_where(self) -> str:
        return f"({self.__column_name},{self.__filter_name},{self.__value})"



import pytest

from .. import filters

from ..nocodb import WhereFilter


@pytest.mark.parametrize('filter_class, expected_operator', [
    (filters.EqFilter, 'eq'),
    (filters.EqualFilter, 'eq'),
    (filters.NotEqualFilter, 'neq'),
    (filters.GreaterOrEqualFilter, 'ge'),
    (filters.GreaterThanFilter, 'gt'),
    (filters.LessThanFilter, 'lt'),
    (filters.LessOrEqualFilter, 'le'),
    (filters.LikeFilter, 'like')
    ])
def test_basic_filters_are_correctly_created(filter_class: WhereFilter, expected_operator: str):
    test_filter = filter_class('column', 'value')
    assert test_filter.get_where() == f'(column,{expected_operator},value)'



from abc import ABC, abstractmethod
from typing import Optional

"""
License MIT

Copyright 2022 Samuel López Saura

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class AuthToken(ABC):
    @abstractmethod
    def get_header(self) -> dict:
        pass


class APIToken:
    def __init__(self, token: str):
        self.__token = token

    def get_header(self) -> dict:
        return {"xc-token": self.__token}


class JWTAuthToken:
    def __init__(self, token: str):
        self.__token = token

    def get_header(self) -> dict:
        return {"xc-auth": self.__token}


class WhereFilter(ABC):
    @abstractmethod
    def get_where(self) -> str:
        pass


"""This could be great but actually I don't know how to join filters in the
NocoDB DSL. I event don't know if this is possible through the current API.
I hope they add docs about it soon.

class NocoDBWhere:

    def __init__(self):
        self.__filter_array: List[WhereFilter] = []

    def add_filter(self, where_filter: WhereFilter) -> NocoDBWhere:
        self.__filter_array.append(
                where_filter
        )
        return self

    def get_where(self) -> str:
        return '&'.join([filter_.get_where() for filter_ in self.__filter_array])

    def __str__(self):
        return f'Where: "{self.get_where()}"'
"""


class NocoDBProject:
    def __init__(self, org_name: str, project_name: str):
        self.project_name = project_name
        self.org_name = org_name


class NocoDBClient:
    @abstractmethod
    def table_row_list(
        self, project: NocoDBProject, table: str, filter_obj=None, params=None
    ) -> dict:
        pass

    @abstractmethod
    def table_row_list(
        self,
        project: NocoDBProject,
        table: str,
        filter_obj: Optional[WhereFilter] = None,
        params: Optional[dict] = None,
    ) -> dict:
        pass

    @abstractmethod
    def table_row_create(
        self, project: NocoDBProject, table: str, body: dict
    ) -> dict:
        pass

    @abstractmethod
    def table_row_detail(
        self, project: NocoDBProject, table: str, row_id: int
    ) -> dict:
        pass

    @abstractmethod
    def table_row_update(
        self, project: NocoDBProject, table: str, row_id: int, body: dict
    ) -> dict:
        pass

    @abstractmethod
    def table_row_delete(
        self, project: NocoDBProject, table: str, row_id: int
    ) -> int:
        pass

    @abstractmethod
    def table_row_nested_relations_list(
        self,
        project: NocoDBProject,
        table: str,
        relation_type: str,
        row_id: int,
        column_name: str,
    ) -> dict:
        pass



from .factory import basic_filter_class_factory, raw_template_filter_class_factory


def test_basic_filter_class_factory():
    FilterClass = basic_filter_class_factory('eq')
    assert FilterClass('column', 'value').get_where() == '(column,eq,value)'


def test_raw_template_filter_class_factory():
    FilterClassWithoutParams = raw_template_filter_class_factory('()')
    FilterClassWithParams = raw_template_filter_class_factory('({},{},{})')
    FilterClassWithKwargs = raw_template_filter_class_factory('({},{op},{})')
    assert FilterClassWithoutParams().get_where() == '()'
    assert FilterClassWithParams('1', '2','3').get_where() == '(1,2,3)'
    assert FilterClassWithKwargs('1', '2', op='eq').get_where() == '(1,eq,2)'



from ..nocodb import WhereFilter

from .basic_filter import BasicFilter
from .raw_filter import RawTemplateFilter


def basic_filter_class_factory(filter_name: str):
    class WrappedFilter(WhereFilter):
        def __init__(self, column_name: str, value: str):
            self.__filter = BasicFilter(column_name, filter_name, value)
        def get_where(self) -> str:
            return self.__filter.get_where()
    return WrappedFilter

def raw_template_filter_class_factory(template: str):
    class WrappedFilter(WhereFilter):
        def __init__(self, *args, **kwargs):
            self.__filter = RawTemplateFilter(template, *args, **kwargs)
        def get_where(self) -> str:
            return self.__filter.get_where()
    return WrappedFilter




from ..nocodb import WhereFilter
from .factory import basic_filter_class_factory


EqFilter = basic_filter_class_factory('eq')
EqualFilter = EqFilter
NotEqualFilter = basic_filter_class_factory('neq')
GreaterThanFilter = basic_filter_class_factory('gt')
GreaterOrEqualFilter = basic_filter_class_factory('ge')
LessThanFilter = basic_filter_class_factory('lt')
LessOrEqualFilter = basic_filter_class_factory('le')
LikeFilter = basic_filter_class_factory('like')



from typing import Optional


class NocoDBAPIError(Exception):
    def __init__(
        self,
        message: str,
        status_code: int,
        response_json: Optional[dict] = None,
        response_text: Optional[str] = None,
    ):
        super().__init__(message)
        self.status_code = status_code
        self.response_json = response_json
        self.response_text = response_text



from enum import Enum
from .nocodb import NocoDBProject


class NocoDBAPIUris(Enum):
    V1_DB_DATA_PREFIX = "api/v1/db/data"
    V1_DB_META_PREFIX = "api/v1/db/meta"


class NocoDBAPI:
    def __init__(self, base_uri: str):
        self.__base_data_uri = (
            f"{base_uri}/{NocoDBAPIUris.V1_DB_DATA_PREFIX.value}"
        )
        self.__base_meta_uri = (
            f"{base_uri}/{NocoDBAPIUris.V1_DB_META_PREFIX.value}"
        )

    def get_table_uri(self, project: NocoDBProject, table: str) -> str:
        return "/".join(
            (
                self.__base_data_uri,
                project.org_name,
                project.project_name,
                table,
            )
        )

    def get_table_count_uri(self, project: NocoDBProject, table: str) -> str:
        return "/".join(
            (
                self.get_table_uri(project, table),
                'count'
            )
        )

    def get_table_find_one_uri(self, project: NocoDBProject, table: str) -> str:
        return "/".join(
            (
                self.get_table_uri(project, table),
                'find-one'
            )
        )

    def get_row_detail_uri(
        self, project: NocoDBProject, table: str, row_id: int
    ):
        return "/".join(
            (
                self.__base_data_uri,
                project.org_name,
                project.project_name,
                table,
                str(row_id),
            )
        )

    def get_nested_relations_rows_list_uri(
        self,
        project: NocoDBProject,
        table: str,
        relation_type: str,
        row_id: int,
        column_name: str,
    ) -> str:
        return "/".join(
            (
                self.__base_data_uri,
                project.org_name,
                project.project_name,
                table,
                str(row_id),
                relation_type,
                column_name,
            )
        )

    def get_project_uri(
        self,
    ) -> str:
        return "/".join(
            (
                self.__base_meta_uri,
                "projects"
            )
        )


from ..nocodb import WhereFilter


class RawFilter(WhereFilter):
    def __init__(self, raw: str):
        self.__raw = raw

    def get_where(self) -> str:
        return self.__raw


class RawTemplateFilter(WhereFilter):
    def __init__(self, template: str, *args, **kwargs):
        self.__template = template
        self.__template_values = args
        self.__template_kvalues = kwargs

    def get_where(self) -> str:
        return self.__template.format(*self.__template_values, **self.__template_kvalues)






from unittest import mock

import pytest

from .requests_client import NocoDBRequestsClient, requests as requests_lib
from ..exceptions import NocoDBAPIError


@mock.patch.object(requests_lib, "Session")
def test_NocoDBAPIError_raised_on_bad_response(mock_requests_session):
    mock_session = mock.Mock()
    expected_exception = NocoDBAPIError("Error 400 in the request", 400)
    mock_session.request.side_effect = expected_exception
    mock_requests_session.return_value = mock_session
    client = NocoDBRequestsClient(mock.Mock(), "")
    with pytest.raises(NocoDBAPIError) as exc_info:
        client._request("GET", "/")

    assert str(exc_info.value) == str(expected_exception)
    assert exc_info.value.status_code == expected_exception.status_code



from setuptools import setup, find_packages


setup(
   name='nocodb',
   version='2.0.0-alpha-1',
   author='Samuel López Saura',
   author_email='samuellopezsaura@gmail.com',
   packages=find_packages(),
   license='MIT',
   url='https://github.com/ElChicoDePython/python-nocodb',
   classifiers=[
       "Programming Language :: Python :: 3",
       "License :: OSI Approved :: MIT License",
       "Operating System :: OS Independent",
   ],
   description='A package to use NocoDB API in a simple way',
   long_description=open('README.md').read(),
   long_description_content_type="text/markdown",
   install_requires=[
       "requests>=2.0",
   ],
)
